import time
from typing import TypeVar, Generic
from abc import ABC, ABCMeta, abstractmethod
import traceback
from functools import reduce

from api.database.database import (PreparedStatement, FilledStatement, DbConnection,
                                   DbResult, DbResultSet, DbValueType, DbAffectedRows,
                                   DatabaseWarning, DatabaseError,
                                   SQL_PARAMETER_INDICATOR, TransactionConflictError)

_DEBUG_PRINT_STATEMENT_EXECUTION: bool = False

Connection = TypeVar("Connection")
Cursor = TypeVar("Cursor")


class PythonDbConnection(DbConnection, Generic[Connection, Cursor], ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self,
                 parameter_indicator: str,
                 caller_exception: tuple[type, ...],
                 py_connection: Connection):
        super().__init__()
        self._parameter_indicator = parameter_indicator
        self._caller_exception = caller_exception
        self._py_connection = py_connection
        self._closed = False
        # str is statement with SQL_PARAMETER_INDICATOR replaced
        self._prepared_statement_cache: dict[PreparedStatement, tuple[str, Cursor or None]] = {}
        self._unprepared_cursor: Cursor or None = None
        self._last_successful_request: int = time.time_ns()
    
    def _clear_cache_on_new_connection(self):
        self._prepared_statement_cache = {}
        self._unprepared_cursor = None
    
    def execute_statements(self, statements: [FilledStatement]) -> list[DbResult]:
        if len(statements) > 1 and self._can_use_multi_query(statements):
            return self._execute_statements_multi_query(statements)
        else:
            return self._execute_statements_prepared(statements)
    
    def _execute_statements_prepared(self, statements: [FilledStatement]) -> list[DbResult]:
        if self._closed:
            raise RuntimeError("Already closed")  # pragma: no cover
        results: list[DbResult] = []
        for filled_stat in statements:
            if _DEBUG_PRINT_STATEMENT_EXECUTION:  # pragma: no cover
                start_time = time.time_ns()
                print(f"Executing statement:\n{filled_stat.statement}\nwith values: {filled_stat.values}")
            
            exception = None
            try:
                if isinstance(filled_stat.statement, PreparedStatement):
                    res = self._execute_prepared_statement(filled_stat.statement, filled_stat.values)
                else:
                    assert isinstance(filled_stat.statement, str)
                    res = self._execute_unprepared_statement(filled_stat.statement, filled_stat.values)
                warnings, affected_rows, result_set = res
            except self._caller_exception as e:  # pragma: no cover
                exception = e
                warnings = []
                affected_rows = None
                result_set = []
            except Exception as e:
                if self._is_transaction_conflict_exception(e):
                    raise TransactionConflictError("Conflict in transactions") from e
                if filled_stat.queue_traceback is not None:
                    raise DatabaseError(f"An exception occurred while executing statement, queued at:\n\n"
                                        f"{reduce(lambda res, s: res + s, traceback.format_list(filled_stat.queue_traceback), '')}") from e
                else:
                    raise DatabaseError("An exception occurred while executing statement") from e
            results.append((exception, warnings, result_set, affected_rows))
            
            if _DEBUG_PRINT_STATEMENT_EXECUTION:  # pragma: no cover
                # noinspection PyUnboundLocalVariable
                print(f"Executed statement in {((time.time_ns() - start_time )//1000 ) /1000} ms\n")
        return results
    
    def _execute_statements_multi_query(self, statements: [FilledStatement]) -> list[DbResult]:
        if self._closed:
            raise RuntimeError("Already closed")

        if _DEBUG_PRINT_STATEMENT_EXECUTION:
            print("Executing multi query with following statements:")
        
        full_query = ""
        all_values = []
        for filled_stat in statements:
            if _DEBUG_PRINT_STATEMENT_EXECUTION:
                print(f"{filled_stat.statement}\nwith values: {filled_stat.values}\n")
            
            if full_query != "":
                full_query += ";"
            if isinstance(filled_stat.statement, PreparedStatement):
                replaced_stat, cursor = self._get_prepared_statement_data(filled_stat.statement)
            else:
                replaced_stat = self._replace_parameter_indicator(filled_stat.statement)
            full_query += replaced_stat
            all_values.extend(filled_stat.values)
        
        if self._unprepared_cursor is None:
            self._unprepared_cursor = self._create_cursor(False)
        
        if _DEBUG_PRINT_STATEMENT_EXECUTION:
            start_time = time.time_ns()
        
        results: list[DbResult] = []
        try:
            for warnings, affected_rows, result_set in self._db_execute_multiple_statements(
                    self._unprepared_cursor, full_query, all_values):
                results.append((None, warnings, result_set, affected_rows))
        except self._caller_exception as e:
            for i in range(0, len(statements)):
                results.append((e, [], [], None, None))
        except Exception as e:
            raise DatabaseError("An exception occurred while executing statements") from e
        
        if _DEBUG_PRINT_STATEMENT_EXECUTION:
            # noinspection PyUnboundLocalVariable
            print(f"Executed query in {((time.time_ns() - start_time) // 1000) / 1000} ms\n")
        return results
    
    def _replace_parameter_indicator(self, statement: str) -> str:
        return statement.replace(SQL_PARAMETER_INDICATOR, self._parameter_indicator)
    
    def _get_prepared_statement_data(self, statement: PreparedStatement):
        if statement in self._prepared_statement_cache:
            return self._prepared_statement_cache[statement]
        replaced_stat = self._replace_parameter_indicator(statement.statement)
        data = (replaced_stat, self._create_cursor(True))
        self._prepared_statement_cache[statement] = data
        return data
    
    def _execute_prepared_statement(self,
                                    statement: PreparedStatement,
                                    values: list[DbValueType]) -> tuple[list[DatabaseWarning], DbAffectedRows, DbResultSet]:
        replaced_stat, cursor = self._get_prepared_statement_data(statement)
        return self._db_execute_single_statement(cursor, replaced_stat, values)
    
    def _execute_unprepared_statement(self,
                                      statement: str,
                                      values: list[DbValueType]) -> tuple[list[DatabaseWarning], DbAffectedRows, DbResultSet]:
        if self._unprepared_cursor is None:
            self._unprepared_cursor = self._create_cursor(False)
        replaced_stat = self._replace_parameter_indicator(statement)
        return self._db_execute_single_statement(self._unprepared_cursor, replaced_stat, values)
    
    @staticmethod
    def _create_column_mapping(cursor: Cursor) -> dict[str, int]:
        column_mapping = {}
        i = -1
        for column_description in cursor.description:
            i += 1
            name = column_description[0]  # As in the Python DB Api v2
            column_mapping[name] = i
        return column_mapping

    @abstractmethod
    def _can_use_multi_query(self, statements: [FilledStatement]):
        pass  # pragma: no cover
    
    @abstractmethod
    def _create_cursor(self, prepared: bool):
        pass  # pragma: no cover
    
    @abstractmethod
    def _is_transaction_conflict_exception(self, exception: Exception) -> bool:
        pass  # pragma: no cover
    
    @abstractmethod
    def _db_execute_single_statement(self,
                                     cursor: Cursor,
                                     statement: str,
                                     values: list[DbValueType]) -> tuple[list[DatabaseWarning], DbAffectedRows, DbResultSet]:
        pass  # pragma: no cover
    
    @abstractmethod
    def _db_execute_multiple_statements(self,
                                        cursor: Cursor,
                                        statements: str,
                                        values: list[DbValueType]) -> list[tuple[list[DatabaseWarning], DbAffectedRows, DbResultSet]]:
        pass  # pragma: no cover
    
    def execute_script(self, script: str):
        if self._closed:
            raise RuntimeError("Already closed")  # pragma: no cover
        
        if _DEBUG_PRINT_STATEMENT_EXECUTION:  # pragma: no cover
            start_time = time.time_ns()
            print(f"Executing script")
        
        try:
            if self._unprepared_cursor is None:
                self._unprepared_cursor = self._create_cursor(False)
            self._db_execute_script(self._unprepared_cursor, script)
        except Exception as e:
            raise DatabaseError("An exception occurred while executing script") from e
        
        if _DEBUG_PRINT_STATEMENT_EXECUTION:  # pragma: no cover
            # noinspection PyUnboundLocalVariable
            print(f"Executed script in {((time.time_ns() - start_time) // 1000) / 1000} ms\n")
    
    @abstractmethod
    def _db_execute_script(self, cursor: Cursor, script: str):
        """
        See execute_script of :class:`DbConnectionPool`
        """
        pass  # pragma: no cover
