from datetime import date, timedelta

from api.routes import *
from api.course import (COURSE_SECONDARY_DB_SELECTION,
                        course_db_to_json_no_lectures, course_secondary_db_to_json_no_lectures,
                        lecture_db_to_json_no_chapters_media, lecture_queue_query_auth,
                        COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE)
from api.announcement import query_announcements

_GIT_COMMIT_HASH = os.environ.get("VIDEOAG_API_GIT_COMMIT_HASH", None)
if _GIT_COMMIT_HASH is not None and len(_GIT_COMMIT_HASH) == 0:
    _GIT_COMMIT_HASH = None

_SERVER_NAME = api.config["API_SERVER_NAME"]


@api_route("/status", ["GET"], allow_while_readonly=True, allow_while_disabled=True)
def api_route_status():
    status = "available"
    announcements = []
    announcements.extend(api.live_config.get_static_announcements())
    
    if api.live_config.is_disabled():
        status = "unavailable"
    else:
        if api.live_config.is_readonly():
            status = "readonly"
        # noinspection PyBroadException
        try:
            announcements.extend(query_announcements())
        except Exception:
            # We do NOT return status 'unavailable' because the status is the EXPECTED status. If there is an unexpected
            # db failure, we want errors to show up and not just 'Site not available'
            pass
    response = {
        "status": status,
        "is_debug": DEBUG_ENABLED,
        "server_name": _SERVER_NAME,
        "announcements": announcements
    }
    if _GIT_COMMIT_HASH is not None:
        response["git_commit_hash"] = _GIT_COMMIT_HASH
    return response


@api_route("/is_running", ["GET"], allow_while_readonly=True, allow_while_disabled=True)
def api_route_is_running():
    return {}, HTTP_200_OK


_SQL_GET_UPCOMING_LECTURES = PreparedStatement(f"""
SELECT "lectures".*, {COURSE_SECONDARY_DB_SELECTION}
FROM "lectures"
JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
WHERE ("time" >= ?)
    AND ("time" <= ?)
    AND ( ? OR (
            "courses"."visible"
        AND "courses"."listed"
        AND "lectures"."visible"
    ))
    AND (NOT "lectures"."norecording")
ORDER BY "time" ASC, "lectures"."id" ASC LIMIT 30
""")
_SQL_GET_UPCOMING_LECTURES_AUTH = PreparedStatement("""
SELECT "perm".* FROM
    (SELECT "lectures"."id" AS "sub_lecture_id", "courses"."id" AS "sub_course_id"
    FROM "lectures"
    JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
    WHERE ("time" >= ?)
        AND ("time" <= ?)
        AND ( ? OR (
                "courses"."visible"
            AND "courses"."listed"
            AND "lectures"."visible"
        ))
        AND (NOT "lectures"."norecording")
    ORDER BY "time" ASC, "lectures"."id" ASC LIMIT 30
    ) as "selected_lectures"
    JOIN "perm" ON ("perm"."lecture_id" = "sub_lecture_id" OR "perm"."course_id" = "sub_course_id")
    WHERE (NOT "perm"."deleted")
""")
_SQL_GET_LATEST_MEDIA = PreparedStatement(f"""
SELECT "lectures".*, {COURSE_SECONDARY_DB_SELECTION}
FROM "lectures"
JOIN (
        SELECT
            "videos"."lecture_id" AS "lecture_id",
            MAX("videos"."time_created") as "time"
        FROM "videos"
        WHERE ? OR "videos"."visible"
        GROUP BY "videos"."lecture_id"
    ) AS "_latest_videos"
    ON ("_latest_videos"."lecture_id" = "lectures"."id")
JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
WHERE ? OR (
            "courses"."visible"
        AND "courses"."listed"
        AND "lectures"."visible"
    )
ORDER BY "_latest_videos"."time" DESC LIMIT 6
""")
_SQL_GET_LATEST_MEDIA_AUTH = PreparedStatement("""
SELECT "perm".* FROM
    (SELECT "lectures"."id" AS "sub_lecture_id", "courses"."id" AS "sub_course_id"
        FROM "lectures"
        JOIN (
                SELECT
                    "videos"."lecture_id" AS "lecture_id",
                    MAX("videos"."time_created") as "time"
                FROM "videos"
                WHERE ? OR "videos"."visible"
                GROUP BY "videos"."lecture_id"
            ) AS "_latest_videos"
            ON ("_latest_videos"."lecture_id" = "lectures"."id")
        JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
        WHERE ? OR (
                    "courses"."visible"
                AND "courses"."listed"
                AND "lectures"."visible"
            )
        ORDER BY "_latest_videos"."time" DESC LIMIT 6
    ) as "selected_lectures"
    JOIN "perm" ON ("perm"."lecture_id" = "sub_lecture_id" OR "perm"."course_id" = "sub_course_id")
    WHERE (NOT "perm"."deleted")
""")
_SQL_GET_ACTIVE_LIVESTREAMS = PreparedStatement(f"""
SELECT "lectures".*, {COURSE_SECONDARY_DB_SELECTION}
FROM "lectures"
JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
WHERE ("lectures"."stream_job" IS NOT NULL)
    AND ( ? OR (
            "courses"."visible"
        AND "lectures"."visible"
        AND "courses"."listed"
        AND "lectures"."visible"
    ))
ORDER BY "lectures"."time" DESC
""")
_SQL_GET_ACTIVE_LIVESTREAMS_AUTH = PreparedStatement("""
SELECT "perm".* FROM
    (SELECT "lectures"."id" AS "sub_lecture_id", "courses"."id" AS "sub_course_id"
    FROM "lectures"
    JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
    WHERE ("lectures"."stream_job" IS NOT NULL)
        AND ( ? OR (
                "courses"."visible"
            AND "lectures"."visible"
            AND "courses"."listed"
            AND "lectures"."visible"
        ))
    ORDER BY "lectures"."time" DESC
    ) as "selected_lectures"
    JOIN "perm" ON ("perm"."lecture_id" = "sub_lecture_id" OR "perm"."course_id" = "sub_course_id")
    WHERE (NOT "perm"."deleted")
""")
_SQL_GET_FEATURED = PreparedStatement("""
SELECT *
FROM "featured"
WHERE (? OR "visible")
    AND (NOT "deleted")
ORDER BY CASE WHEN "order" IS NULL THEN 0 ELSE "order" END ASC, "id" ASC
""")


def _db_execute_get_homepage(transaction: ReadTransaction, is_mod: bool, upcoming_start: date, upcoming_end: date):
    upcoming_db, \
        upcoming_auth_db, \
        latest_media_db, \
        latest_media_auth_db, \
        active_livestreams_db, \
        active_livestreams_auth_db, \
        featured_list_db \
        = transaction.execute_statements(
            (_SQL_GET_UPCOMING_LECTURES, [upcoming_start, upcoming_end, is_mod]),
            (_SQL_GET_UPCOMING_LECTURES_AUTH, [upcoming_start, upcoming_end, is_mod]),
            (_SQL_GET_LATEST_MEDIA, [is_mod, is_mod]),
            (_SQL_GET_LATEST_MEDIA_AUTH, [is_mod, is_mod]),
            (_SQL_GET_ACTIVE_LIVESTREAMS, [is_mod]),
            (_SQL_GET_ACTIVE_LIVESTREAMS_AUTH, [is_mod]),
            (_SQL_GET_FEATURED, [is_mod])
        )
    courses_context = {}
    featured = get_homepage_featured(transaction, courses_context, is_mod, featured_list_db)
    # Transaction is closed by get_homepage_featured
    
    return {
        "featured": featured,
        "upcoming_lectures": get_homepage_upcoming_lectures(
            courses_context, is_mod,
            upcoming_db, upcoming_auth_db
        ),
        "new_lectures": get_homepage_latest_lectures(
            courses_context, is_mod,
            latest_media_db, latest_media_auth_db,
            active_livestreams_db, active_livestreams_auth_db
        ),
        "courses_context": courses_context
    }


@api_route("/homepage", ["GET"], allow_while_readonly=True)
def api_route_homepage():
    is_mod: bool = is_moderator()
    upcoming_start: date = date.today()
    upcoming_end: date = upcoming_start + timedelta(days=7)
    
    return db_pool.execute_read_transaction(
        _db_execute_get_homepage,
        is_mod,
        upcoming_start,
        upcoming_end
    )


def get_homepage_upcoming_lectures(
        courses_context: {},
        is_mod: bool,
        upcoming_db: DbResultSet,
        upcoming_auth_db: DbResultSet):
    upcoming_auth_i = 0
    upcoming_json = []
    for lecture_db in upcoming_db:
        upcoming_auth_i, course_auth_db = db_collect_id_sorted_data(
            upcoming_auth_db, upcoming_auth_i,
            "course_id", lecture_db["course_id"])
        upcoming_auth_i, lecture_auth_db = db_collect_id_sorted_data(
            upcoming_auth_db, upcoming_auth_i,
            "lecture_id", lecture_db["id"])
        upcoming_json.append(lecture_db_to_json_no_chapters_media(
            lecture_db, course_auth_db + lecture_auth_db,
            not bool(lecture_db[COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE]), is_mod))
        
        if str(lecture_db["course_id"]) not in courses_context:
            courses_context[str(lecture_db["course_id"])] = \
                course_secondary_db_to_json_no_lectures(lecture_db, course_auth_db, is_mod)
    
    return upcoming_json


def get_homepage_latest_lectures(
        courses_context: {},
        is_mod: bool,
        latest_media_db: DbResultSet,
        latest_media_auth_db: DbResultSet,
        active_livestreams_db: DbResultSet,
        active_livestreams_auth_db: DbResultSet):
    
    new_lectures_json = []
    
    def add_to_lectures(lectures_db, lectures_auth_db):
        for lecture_db in lectures_db:
            lecture_auth_db = db_collect_unsorted_data(lectures_auth_db, "lecture_id", lecture_db["id"])
            course_auth_db = db_collect_unsorted_data(lectures_auth_db, "course_id", lecture_db["course_id"])
            new_lectures_json.append(lecture_db_to_json_no_chapters_media(
                lecture_db, course_auth_db + lecture_auth_db,
                not bool(lecture_db[COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE]), is_mod))
            
            if str(lecture_db["course_id"]) not in courses_context:
                courses_context[str(lecture_db["course_id"])] = \
                    course_secondary_db_to_json_no_lectures(lecture_db, course_auth_db, is_mod)
    
    add_to_lectures(active_livestreams_db, active_livestreams_auth_db)
    add_to_lectures(latest_media_db, latest_media_auth_db)
    
    return new_lectures_json


_SQL_GET_FEATURED_LECTURES = PreparedStatement(f"""
SELECT "lectures".*, {COURSE_SECONDARY_DB_SELECTION}
FROM "lectures"
JOIN "courses" ON ("lectures"."course_id" = "courses"."id")
WHERE "lectures"."id" = ?
    AND "courses"."visible"
    AND "lectures"."visible"
""")
_SQL_GET_FEATURED_COURSES = {order_type: PreparedStatement(f"""
SELECT * FROM "courses"
WHERE "visible"
    AND "listed"
    AND "{order_type}" = ?
ORDER BY "id" ASC
""") for order_type in ["title", "semester", "organizer", "subject"]}
_SQL_GET_FEATURED_COURSES_AUTH = {order_type: PreparedStatement(f"""
SELECT "perm".* FROM "perm"
JOIN "courses" ON ("courses"."id" = "perm"."course_id")
WHERE "courses"."visible"
    AND "courses"."listed"
    AND "courses"."{order_type}" = ?
    AND (NOT "perm"."deleted")
ORDER BY "courses"."id" ASC
""") for order_type in _SQL_GET_FEATURED_COURSES.keys()}


def get_homepage_featured(
        transaction: ReadTransaction,
        courses_context: {},
        is_mod: bool,
        featured_list_db: DbResultSet):
    featured_list_json = []
    # Row is lecture featured, first statement is for lecture, second statement is for auth
    queued_video_features: list[tuple[dict, FilledStatement, FilledStatement]] = []
    # Row is courses featured, first statement is for courses, second statement is for auth
    queued_courses_features: list[tuple[dict, FilledStatement, FilledStatement]] = []
    feature_display_prio = 0
    for featured_db in featured_list_db:
        featured_json = {
            "title": featured_db["title"],
            "text": featured_db["text"]
        }
        if featured_db["type"] == "plain":
            featured_json["type"] = "plain"
        elif featured_db["type"] == "image":
            featured_json["type"] = "image"
            featured_json["image_url"] = featured_db["param"]
        elif featured_db["type"] == "video":
            featured_json["type"] = "lecture"
            lecture_id = int(featured_db["param"])
            queued_video_features.append((
                featured_json,
                transaction.queue_statement(_SQL_GET_FEATURED_LECTURES, lecture_id),
                lecture_queue_query_auth(transaction, lecture_id)
            ))
        elif featured_db["type"] == "courses":
            featured_json["type"] = "course_list"
            order_type = featured_db["param"]
            if order_type not in _SQL_GET_FEATURED_COURSES:
                featured_json["courses"] = []
                continue
            queued_courses_features.append((
                featured_json,
                transaction.queue_statement(_SQL_GET_FEATURED_COURSES[order_type], featured_db["param2"]),
                transaction.queue_statement(_SQL_GET_FEATURED_COURSES_AUTH[order_type], featured_db["param2"])
            ))
        else:
            raise Exception("Unknown featured type: " + str(featured_db["type"]))  # pragma: no cover
        if is_mod:
            featured_json["id"] = featured_db["id"]
            featured_json["is_visible"] = bool(featured_db["visible"])
            featured_json["display_priority"] = feature_display_prio
        feature_display_prio += 1
        featured_list_json.append(featured_json)
    
    transaction.close()
    
    for featured_json, lecture_stat, lecture_auth_stat in queued_video_features:
        lecture_list_db = transaction.get_result(lecture_stat)
        lecture_auth_db = transaction.get_result(lecture_auth_stat)
        if len(lecture_list_db) == 0:
            featured_json["lecture"] = None
            continue
        lecture_db = lecture_list_db[0]
        course_id = lecture_db["course_id"]
        featured_json["lecture"] = lecture_db_to_json_no_chapters_media(
            lecture_db, lecture_auth_db, not bool(lecture_db[COURSE_SECONDARY_COLUMN_EMBED_INVISIBLE]), is_mod)
        
        if str(course_id) not in courses_context:
            courses_context[str(course_id)] = course_secondary_db_to_json_no_lectures(
                lecture_db,
                lecture_auth_db,
                is_mod)
    
    for featured_json, courses_stat, courses_auth_stat in queued_courses_features:
        courses_db = transaction.get_result(courses_stat)
        courses_auth_db = transaction.get_result(courses_auth_stat)
        
        courses_id_list: list[int] = []
        
        courses_auth_db_i = 0
        for course_db in courses_db:
            id: int = course_db["id"]
            courses_id_list.append(id)
            courses_auth_db_i, course_auth_db = db_collect_id_sorted_data(
                courses_auth_db,
                courses_auth_db_i,
                "course_id",
                id
            )
            if str(id) not in courses_context:
                courses_context[str(id)] = course_db_to_json_no_lectures(course_db, course_auth_db, is_mod)
        
        featured_json["courses"] = courses_id_list
    
    return featured_list_json
