from abc import ABC, ABCMeta, abstractmethod

from api.miscellaneous import *
from api.database import *


class SettingsEntry(ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self, id: str, type_id: str):
        super().__init__()
        self.id = id
        self._type_id = type_id
    
    def to_json(self, settings_db: DbResultRow) -> dict:
        return {
            "id": self.id,
            "type": self._type_id
        }


class SettingsValueEntry(SettingsEntry, ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self, id: str, type_id: str, db_column: str):
        super().__init__(id, type_id)
        self.db_column = db_ensure_valid_id(db_column)
    
    def to_json(self, settings_db: DbResultRow) -> dict:
        return super().to_json(settings_db) | {
            "value": self._db_value_to_json(settings_db[self.db_column])
        }
    
    @abstractmethod
    def _db_value_to_json(self, db_value: DbValueType):
        pass  # pragma: no cover
    
    @abstractmethod
    def json_value_to_db(self, json_value: CJsonValue):
        pass  # pragma: no cover


class SettingsBooleanEntry(SettingsValueEntry):
    
    def __init__(self, id: str, db_colum: str):
        super().__init__(id, "boolean", db_colum)
    
    def _db_value_to_json(self, db_value: DbValueType):
        return bool(db_value)
    
    def json_value_to_db(self, json_value: CJsonValue):
        return json_value.as_bool()


class SettingsSubSectionEntry(SettingsEntry):
    
    def __init__(self, id: str, entries: list[SettingsValueEntry]):
        super().__init__(id, "subsection")
        self.entries = entries
    
    def to_json(self, settings_db: DbResultRow) -> dict:
        return super().to_json(settings_db) | {
            "entries": list(map(lambda e: e.to_json(settings_db), self.entries))
        }


class SettingsSection:
    
    def __init__(self, name_key: str, entries: list[SettingsEntry]):
        super().__init__()
        self._name_key = name_key
        self.entries = entries
    
    def to_json(self, settings_db: DbResultRow):
        return {
            "name": self._name_key,
            "entries": list(map(lambda e: e.to_json(settings_db), self.entries))
        }


class Settings:
    
    def __init__(self, sections: list[SettingsSection]):
        super().__init__()
        self._sections = sections
        self._value_entries_by_id: dict[str, SettingsValueEntry] = {}
        for section in sections:
            self._collect_entries(section.entries)
    
    def _collect_entries(self, entries: list[SettingsEntry]):
        for entry in entries:
            if isinstance(entry, SettingsSubSectionEntry):
                self._collect_entries(entry.entries)
                continue
            assert isinstance(entry, SettingsValueEntry)
            if entry.id in self._value_entries_by_id:
                raise RuntimeError(f"Duplicate settings id {entry.id}")  # pragma: no cover
            self._value_entries_by_id[entry.id] = entry
    
    def get_value_entry(self, id: str) -> SettingsValueEntry or None:
        return self._value_entries_by_id.get(id, None)
    
    def to_json(self, settings_db: DbResultRow):
        return {
            "sections": list(map(lambda s: s.to_json(settings_db), self._sections))
        }
